/****************************************************************************
 *  RES2CLIP.cc                                                       v1.2  *
 *                                                                          *
 *  Function:    Maps *o.resref files to expert system (CLIPS) format       *
 *  Compiler:    g++                                                        *
 *  System:      Digitial VXT2000+                                          *
 *  Application: NPAC/NMPC Transient Stability Analysis                     *
 *               Expert System Component                                    *
 *  Author:      Siva G. Narendra, NPAC                                     *
 *                                                                          *
 *              (C) Northeast Parallel Architectures Center                 *
 *                  Syracuse University, Syracuse, NY                       *
 ****************************************************************************


	DESCRIPTION:
		res2clip  M B F G <f1> <f2> ... <fn> <outputFile>
   
		M (integer) represents the MW o/p of the generators
		for which <f1>..<fn> describes the plot
		of n different parameters. (MANDATORY)

		B (integer) represents the bucket size in samples per
		bucket. (MANDATORY)

		If F=max then MAX() will be used to transform
		the B samples in a bucket to 1 sample.
		If F=mean then MEAN() will be used to do the
		same. (MANDATORY)

		If G=y, then xgraph will be plotted, if G=n then
		the graph will not be plotted (MANDATORY).

		<f1> <f2> ... <fn>
		Represents the n input (*o.resref) files.
		The order of the input files signifies the
		order of analysis required at the expert
		system level. We assume there is 1 parameter
		per *o.resref file. (ATLEAST 1 i/p FILE IS 
		MANDATORY)

		<outputFile>
		Name of the file were the CLIPS data is to
		be stored. (MANDATORY)

	In v1.2 is it assumed that there will be 1 parameter plot
	in each i/p file and that all the i/p files represent the status
	of a paramater for 'M' MW power generation.
	All fields marked MANDATORY, SHOULD be present in the 
	command-line. 
	The ORDER of fields in the command-line SHOULD be as above.





	HISTORY:
		- 04 April 1994 - v1.0
		  created the basic skeleton of res2clip utility
		- 06 April 1994 - v1.1
		  added xgraph display option and tested conversion
		  for sample test cases
	    - 08 April 1994 - v1.2
		  the code was changed to read plot files of OAP 
		  versions 3.0 and 3.1 

**************************************************************************/

#define LINE_MAX 512                  // Max. no. of char./line
#define B_MIN 4                       // Min. no. of samples/bucket
#define B_MAX 15                      // Max. no. of samples/bucket
#define PP_MAX 200                    // Max. no. of pk2pk cycles/waveform
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static int mstatus;                   // The i/ps are for mstatusMW
static int bstatus;                   // Samples per bucket
static int fstatus;                   // 0=>MAX(), 1=>MEAN()
static int Index;                     // Measure of total pk2pk samples
static int repeat;                    // Samples after transformation
static float *Write;
static float *value;                  // Arrays used to store plots



void syntax (int count, char *line0, char *line1, char *line2, char *line3);
void printUsage (char *utility);
void mean ( );
void max ( );





void syntax (int count, char *line0, char *line1, char *line2, char *line3) {
	if (count<7) {
		printUsage(line0);
		exit (0);
	}
	mstatus=atoi (line1);
	bstatus=atoi (line2);
	if (strcmp(line3, "mean")) fstatus=0;
	else fstatus=1;
	if (mstatus <0) {
		cerr<<"\nERROR: Generated Power cannot be Negative!\n\n";
		exit(0);
	}
	if (bstatus<B_MIN) bstatus=B_MIN;
	if (bstatus>B_MAX) bstatus=B_MAX;
}






void printUsage (char *utility) {
    cerr<<"\n\nUSAGE (v1.2):\n";
    cerr<<"    res2clip  M B F G <f1> <f2> ... <fn> <outputFile>\n\n";

    cerr<<"    M (integer) represents the MW o/p of the generators\n"; 
    cerr<<"    for which <f1>..<fn> describes the plot\n";
    cerr<<"    of n different parameters. (MANDATORY)\n\n";

    cerr<<"    B (integer) represents the bucket size in samples per\n";
    cerr<<"    bucket. MAX/MIN: "<<B_MAX<<"/"<<B_MIN<<"  (MANDATORY)\n\n";

    cerr<<"    If F=max then MAX() will be used to transform\n";
    cerr<<"    the B samples in a bucket to 1 sample.\n";
    cerr<<"    If F=mean then MEAN() will be used to do the\n";
    cerr<<"    same. (MANDATORY)\n\n";

    cerr<<"    If G=y, then xgraph will be plotted, if G=n then\n";
    cerr<<"    the graph will not be plotted (MANDATORY).\n\n";

    cerr<<"    <f1> <f2> ... <fn>\n";
    cerr<<"    Represents the n input plot files (of OAP v3.0/v3.1 format).\n";
    cerr<<"    The order of the input files signifies the\n";
    cerr<<"    order of analysis required at the expert\n";
    cerr<<"    system level. We assume there is 1 parameter\n";
    cerr<<"    per *o.resref file. (ATLEAST 1 i/p FILE IS\n";
    cerr<<"    MANDATORY)\n\n";

    cerr<<"    <outputFile>\n";
    cerr<<"    Name of the file were the CLIPS data is to\n";
    cerr<<"    be stored. (MANDATORY)\n\n";

    cerr<<"In v1.2 is it assumed that there will be 1 parameter plot\n";
    cerr<<"in each i/p file and that all the i/p files represent the status\n";
    cerr<<"of a paramater for 'M' MW power generation.\n";
    cerr<<"All fields marked MANDATORY, SHOULD be present in the\n";
    cerr<<"command-line. Atleast 1 i/p file should be given.\n";
    cerr<<"The ORDER of fields in the command-line SHOULD be as above.\n\n";

}







void mean ( ) {
	float sum=0;
	int toCalc=(Index-1)/2;
	int residue=toCalc%bstatus;

	repeat=toCalc/bstatus;
	value=new float [repeat];
	for (int i=0; i<repeat; i++) {
		sum=0;
		for (int j=0; j<bstatus; j++) {
			sum+=Write [i*bstatus+j];
		}
		value[i]=sum/bstatus;
	}
	sum=0;
  	for (int j=0; j<residue; j++){
		sum+=Write[repeat*bstatus+j];	
	}
	value[repeat]=sum/bstatus;	
}





void max ( ) {
    float max;
    int toCalc=(Index-1)/2;
    int residue=toCalc%bstatus;

    repeat=toCalc/bstatus;
    value=new float [repeat];
    for (int i=0; i<repeat; i++) {
        max=0;
        for (int j=0; j<bstatus; j++) {
            float temp=Write [i*bstatus+j];
            if (temp>max) max=temp;           
        }
        value[i]=max;
    }
    max=0;
    for (int j=0; j<residue; j++){  
        float temp=Write[repeat*bstatus+j];
        if (temp>max) max=temp;
    }
    value[repeat]=max;
}







int main (int argc, char *argv[]) {


	syntax (argc, argv[0], argv[1], argv[2], argv[3]);
	cout<<"\n\nPower Output:            "<<mstatus<<"MW\n";
	cout<<"Bucket Size:             "<<bstatus;
	cout<<" Samples/bucket (Min/Max: "<<B_MIN<<"/"<<B_MAX<<")\n";
	cout<<"Transformation Function: ";
	if (fstatus) cout<<"mean()\n";
	else cout<<"max()\n";


	cout<<"\n...trying to open "<<argv[argc-1]<<" for output\n";
	ofstream output(argv[argc-1]);
	if (!output) {
		cerr<<"\nERROR: Unable to open "<<argv[argc-1];
		cerr<<" for output\n"<<"       (file already exists?)\n\n";
		exit (0);
	}


	output<<"(deffacts MW_"<<mstatus<<"_DATA\n";
	output<<"\n(MW "<<mstatus<<")\n";
	output<<"(BUCKET SIZE "<<bstatus<<")\n";
	output<<"(TRANSFORMATION FUNCTION ";
	if (fstatus==0) output<<"MAX)\n";
	else output<<"MEAN)\n";


	cout<<"\n...trying to open graph.r2c for storing plot\n";	
	ofstream graph("graph.r2c");
	if (!graph) {
		cerr<<"WARNING: Unable to open graph.r2c for output\n";
	}


	int test, priority, points, goingDown, first;
	int start, terminate;						//added for OAP 3.1
	float dummy, prev, curr;
	float *record;
	char  *buffer;
	priority=0;
	record=new float [PP_MAX];
	Write =new float [PP_MAX];
	buffer=new char  [LINE_MAX];



	for (int i=1; i<=argc-6; i++) {             //for each input file
		if (!strcmp(argv[4+i], argv[argc-1])) {
			cout<<"\nERROR: Cannot open ";
			cout<<argv[4+i]<<" for input & output\n\n";
			exit (0);
		} 
		cout<<"\n...trying to open "<<argv[4+i]<<" for input\n";
		ifstream input(argv[4+i]);
		if (!input) {
			cout<<"WARNING: Unable to open "<<argv[4+i];
			cout<<" for input...skipping...\n";
		}
		else {
			Index=0;
			test=0;
			goingDown=-1;
			first=1;
			priority++;

			input>>buffer;
			if (buffer[0]=='@') {				//OAP v3.0
				do {
					if (test++>100)  {
						cout<<"\nERROR: Unexpectd i/p file format\n\n";
						exit(0);
					}
					input>>buffer;
				} while (buffer[0]!='1');

				input>>buffer;
				if (buffer[0]!='=')	{
					cout<<"\nERROR: Unexpectd i/p file format\n\n";
   	                exit(0);
   		        }
				graph<<"\""<<mstatus;
				output<<"\n(PARAMETER NAME ";

				do {
					if (!first){
						output<<"_"<<buffer;
					}
					else  {
						output<<buffer;
						first=0;
					}
					input>>buffer;
				} while (buffer[0]!='(' && buffer[0]!='Y');
				output<<" PRIORITY "<<priority<<")\n";
				graph<<"_"<<priority<<"\n";
	
				do {
					input>>buffer;
				} while (buffer[0]!='X');
				input>>buffer;
				input>>points;
				input>>dummy>>prev;
			}

			else {									//OAP v3.1
				do {
					if (test++>20)  {
						cout<<"\nERROR: Unexpectd i/p file format\n\n";
						exit(0);
					}
					input>>buffer;
				} while (buffer[0]!='C' || buffer[1]!='H');

				input>>buffer;   input>>buffer;
				graph<<"\""<<mstatus;
				output<<"\n(PARAMETER NAME ";

				do {
					if (!first){
						output<<"_"<<buffer;
					}
					else  {
						output<<buffer;
						first=0;
					}
					input>>buffer;
				} while (buffer[0]!='(' && (buffer[0]!='C'||buffer[1]!='H'));
				output<<" PRIORITY "<<priority<<")\n";
				graph<<"_"<<priority<<"\n";

				if (!strcmp(buffer,"CH#")) {    	//buffer==CH#
					input>>buffer;   input>>buffer;
				}
				else {
					do {
						input>>buffer;
					} while ((buffer[0]!='C'||buffer[1]!='H'));
					input>>buffer;   input>>buffer;		
				} 
						
				points=32767;					    //maximum (infinity!)
				input>>dummy>>prev;						
			}


			start=1;	   //allow zero slope in the begining (OAP v3.1)
			terminate=0;   //terminate flag for zero sloe at the end (OAP v3.1) 

			for (int j=2; j<=points; j++) {                     //read all the points
				input>>dummy>>curr;
				if ((curr==prev)&&(start==1)) continue;		    //start
				else if ((curr==prev)&&(terminate++>10)) break;	//end
				if (curr>prev && goingDown!=0) {				
					start=0;  terminate=0;		
					if (goingDown==-1) goingDown=0;
					else {
						record[Index++]=prev;
						graph<<Index<<" "<<prev<<"\n";
						goingDown=0;
					}
				}
				if (curr<prev && goingDown!=1) {
					start=0;  terminate=0;
					if (goingDown==-1) goingDown=1;
					else {
						record[Index++]=prev;
						graph<<Index<<" "<<prev<<"\n";
						goingDown=1;
					}
				}
				prev=curr;
			}



			for (j=0; j<=(Index-1)/2; j++) {          //pk2pk plot
				Write[j]=record[j*2]-record[j*2+1];
				if (Write[j]<0) Write[j]*=-1;
			}



			if (fstatus==0) max();                   //Bsmpls->1smpl
			else mean();


			for (int k=0; k<=repeat; k++) {
				output<<"(PRIORITY "<<priority<<" BUCKET NUMBER ";
				output<<k<<" VALUE "<<value[k]<<")\n";	
			}
			graph<<"\n";


		}
		input.close();
	}
	output<<"\n)\n";
	if (strcmp(argv[4], "n")) 
		system (" xgraph -bb -tk -t RES2CLIP graph.r2c &");
	graph.close();
	output.close();
	cout<<"\n";
	delete record, buffer, Write, value;
	return 1;
}



